from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from stacker.session_cache import get_session
import re
import operator

from . import LookupHandler
from ...util import read_value_from_path

TYPE_NAME = "ami"


class ImageNotFound(Exception):
    def __init__(self, search_string):
        self.search_string = search_string
        message = ("Unable to find ec2 image with search string: {}").format(
            search_string
        )
        super(ImageNotFound, self).__init__(message)


class AmiLookup(LookupHandler):
    @classmethod
    def handle(cls, value, provider, **kwargs):
        """Fetch the most recent AMI Id using a filter
    
        For example:
    
            ${ami [<region>@]owners:self,account,amazon name_regex:serverX-[0-9]+ architecture:x64,i386}
    
            The above fetches the most recent AMI where owner is self
            account or amazon and the ami name matches the regex described,
            the architecture will be either x64 or i386
    
            You can also optionally specify the region in which to perform the
            AMI lookup.
    
            Valid arguments:
    
            owners (comma delimited) REQUIRED ONCE:
                aws_account_id | amazon | self
    
            name_regex (a regex) REQUIRED ONCE:
                e.g. my-ubuntu-server-[0-9]+
    
            executable_users (comma delimited) OPTIONAL ONCE:
                aws_account_id | amazon | self
    
            Any other arguments specified are sent as filters to the aws api
            For example, "architecture:x86_64" will add a filter
        """  # noqa
        value = read_value_from_path(value)

        if "@" in value:
            region, value = value.split("@", 1)
        else:
            region = provider.region

        ec2 = get_session(region).client('ec2')

        values = {}
        describe_args = {}

        # now find any other arguments that can be filters
        matches = re.findall('([0-9a-zA-z_-]+:[^\s$]+)', value)
        for match in matches:
            k, v = match.split(':', 1)
            values[k] = v

        if not values.get('owners'):
            raise Exception("'owners' value required when using ami")
        owners = values.pop('owners').split(',')
        describe_args["Owners"] = owners

        if not values.get('name_regex'):
            raise Exception("'name_regex' value required when using ami")
        name_regex = values.pop('name_regex')

        executable_users = None
        if values.get('executable_users'):
            executable_users = values.pop('executable_users').split(',')
            describe_args["ExecutableUsers"] = executable_users

        filters = []
        for k, v in values.items():
            filters.append({"Name": k, "Values": v.split(',')})
        describe_args["Filters"] = filters

        result = ec2.describe_images(**describe_args)

        images = sorted(result['Images'],
                        key=operator.itemgetter('CreationDate'),
                        reverse=True)
        for image in images:
            if re.match("^%s$" % name_regex, image['Name']):
                return image['ImageId']

        raise ImageNotFound(value)



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from . import LookupHandler


TYPE_NAME = "hook_data"


class HookDataLookup(LookupHandler):
    @classmethod
    def handle(cls, value, context, **kwargs):
        """Returns the value of a key for a given hook in hook_data.

        Format of value:

            <hook_name>::<key>
        """
        try:
            hook_name, key = value.split("::")
        except ValueError:
            raise ValueError("Invalid value for hook_data: %s. Must be in "
                             "<hook_name>::<key> format." % value)

        return context.hook_data[hook_name][key]



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import bytes, str

import base64
import json
import re
try:
    from collections.abc import Mapping, Sequence
except ImportError:
    from collections import Mapping, Sequence

import yaml

from troposphere import GenericHelperFn, Base64

from . import LookupHandler
from ...util import read_value_from_path


TYPE_NAME = "file"

_PARAMETER_PATTERN = re.compile(r'{{([::|\w]+)}}')


class FileLookup(LookupHandler):
    @classmethod
    def handle(cls, value, **kwargs):
        """Translate a filename into the file contents.

        Fields should use the following format::

            <codec>:<path>

        For example::

            # We've written a file to /some/path:
            $ echo "hello there" > /some/path

            # In stacker we would reference the contents of this file with the
            # following
            conf_key: ${file plain:file://some/path}

            # The above would resolve to
            conf_key: hello there

            # Or, if we used wanted a base64 encoded copy of the file data
            conf_key: ${file base64:file://some/path}

            # The above would resolve to
            conf_key: aGVsbG8gdGhlcmUK

        Supported codecs:

            - plain

            - base64 - encode the plain text file at the given path with base64
              prior to returning it

            - parameterized - the same as plain, but additionally supports
              referencing template parameters to create userdata that's
              supplemented with information from the template, as is commonly
              needed in EC2 UserData. For example, given a template parameter
              of BucketName, the file could contain the following text::

                #!/bin/sh
                aws s3 sync s3://{{BucketName}}/somepath /somepath

              and then you could use something like this in the YAML config
              file::

                UserData: ${file parameterized:/path/to/file}

              resulting in the UserData parameter being defined as::

                  { "Fn::Join" : ["", [
                      "#!/bin/sh\\naws s3 sync s3://",
                      {"Ref" : "BucketName"},
                      "/somepath /somepath"
                  ]] }

            - parameterized-b64 - the same as parameterized, with the results
              additionally wrapped in *{ "Fn::Base64": ... }* , which is what
              you actually need for EC2 UserData

        When using parameterized-b64 for UserData, you should use a variable
        defined as such:

        .. code-block:: python

            from troposphere import AWSHelperFn

              "UserData": {
                  "type": AWSHelperFn,
                  "description": "Instance user data",
                  "default": Ref("AWS::NoValue")
              }

        and then assign UserData in a LaunchConfiguration or Instance to
        *self.get_variables()["UserData"]*. Note that we use AWSHelperFn as the
        type because the parameterized-b64 codec returns either a Base64 or a
        GenericHelperFn troposphere object
        """

        try:
            codec, path = value.split(":", 1)
        except ValueError:
            raise TypeError(
                "File value must be of the format"
                " \"<codec>:<path>\" (got %s)" % (value)
            )

        value = read_value_from_path(path)

        return CODECS[codec](value)


def _parameterize_string(raw):
    """Substitute placeholders in a string using CloudFormation references

    Args:
        raw (`str`): String to be processed. Byte strings are not
        supported; decode them before passing them to this function.

    Returns:
        `str` | :class:`troposphere.GenericHelperFn`: An expression with
            placeholders from the input replaced, suitable to be passed to
            Troposphere to be included in CloudFormation template. This will
            be the input string without modification if no substitutions are
            found, and a composition of CloudFormation calls otherwise.
    """

    parts = []
    s_index = 0

    for match in _PARAMETER_PATTERN.finditer(raw):
        parts.append(raw[s_index:match.start()])
        parts.append({u"Ref": match.group(1)})
        s_index = match.end()

    if not parts:
        return GenericHelperFn(raw)

    parts.append(raw[s_index:])
    return GenericHelperFn({u"Fn::Join": [u"", parts]})


def parameterized_codec(raw, b64):
    """Parameterize a string, possibly encoding it as Base64 afterwards

    Args:
        raw (`str` | `bytes`): String to be processed. Byte strings will be
            interpreted as UTF-8.
        b64 (`bool`): Whether to wrap the output in a Base64 CloudFormation
            call

    Returns:
        :class:`troposphere.AWSHelperFn`: output to be included in a
        CloudFormation template.
    """

    if isinstance(raw, bytes):
        raw = raw.decode('utf-8')

    result = _parameterize_string(raw)

    # Note, since we want a raw JSON object (not a string) output in the
    # template, we wrap the result in GenericHelperFn (not needed if we're
    # using Base64)
    return Base64(result.data) if b64 else result


def _parameterize_obj(obj):
    """Recursively parameterize all strings contained in an object.

    Parameterizes all values of a Mapping, all items of a Sequence, an
    unicode string, or pass other objects through unmodified.

    Byte strings will be interpreted as UTF-8.

    Args:
        obj: data to parameterize

    Return:
        A parameterized object to be included in a CloudFormation template.
        Mappings are converted to `dict`, Sequences are converted to  `list`,
        and strings possibly replaced by compositions of function calls.
    """

    if isinstance(obj, Mapping):
        return dict((key, _parameterize_obj(value))
                    for key, value in obj.items())
    elif isinstance(obj, bytes):
        return _parameterize_string(obj.decode('utf8'))
    elif isinstance(obj, str):
        return _parameterize_string(obj)
    elif isinstance(obj, Sequence):
        return list(_parameterize_obj(item) for item in obj)
    else:
        return obj


class SafeUnicodeLoader(yaml.SafeLoader):
    def construct_yaml_str(self, node):
        return self.construct_scalar(node)


def yaml_codec(raw, parameterized=False):
    data = yaml.load(raw, Loader=SafeUnicodeLoader)
    return _parameterize_obj(data) if parameterized else data


def json_codec(raw, parameterized=False):
    data = json.loads(raw)
    return _parameterize_obj(data) if parameterized else data


CODECS = {
    "plain": lambda x: x,
    "base64": lambda x: base64.b64encode(x.encode('utf8')),
    "parameterized": lambda x: parameterized_codec(x, False),
    "parameterized-b64": lambda x: parameterized_codec(x, True),
    "yaml": lambda x: yaml_codec(x, parameterized=False),
    "yaml-parameterized": lambda x: yaml_codec(x, parameterized=True),
    "json": lambda x: json_codec(x, parameterized=False),
    "json-parameterized": lambda x: json_codec(x, parameterized=True),
}



"""Cloudformation module."""

import logging
import os
import platform
import re
import sys

import yaml

from . import RunwayModule, run_module_command
from ..util import change_dir, get_embedded_lib_path

LOGGER = logging.getLogger('runway')


def ensure_stacker_compat_config(config_filename):
    """Ensure config file can be loaded by Stacker."""
    try:
        with open(config_filename, 'r') as stream:
            yaml.load(stream)
    except yaml.constructor.ConstructorError as yaml_error:
        if yaml_error.problem.startswith(
                'could not determine a constructor for the tag \'!'):
            LOGGER.error('"%s" appears to be a CloudFormation template, '
                         'but is located in the top level of a module '
                         'alongside the CloudFormation config files (i.e. '
                         'the file or files indicating the stack names & '
                         'parameters). Please move the template to a '
                         'subdirectory.',
                         config_filename)
            sys.exit(1)


def gen_stacker_env_files(environment, region):
    """Generate possible Stacker environment filenames."""
    return [
        # Give preference to explicit environment-region files
        "%s-%s.env" % (environment, region),
        # Fallback to environment name only
        "%s.env" % environment
    ]


def get_stacker_env_file(path, environment, region):
    """Determine Stacker environment file name."""
    for name in gen_stacker_env_files(environment, region):
        if os.path.isfile(os.path.join(path, name)):
            return name
    return "%s-%s.env" % (environment, region)  # fallback to env & region


def make_stacker_cmd_string(args, lib_path):
    """Generate stacker invocation script from command line arg list.

    This is the standard stacker invocation script, with the following changes:
    * Adding our explicit arguments to parse_args (instead of leaving it empty)
    * Overriding sys.argv
    * Adding embedded runway lib directory to sys.path
    """
    if platform.system().lower() == 'windows':
        # Because this will be run via subprocess, the backslashes on Windows
        # will cause command errors
        lib_path = lib_path.replace('\\', '/')
    return ("import sys;"
            "sys.argv = ['stacker'] + {args};"
            "sys.path.insert(1, '{lib_path}');"
            "from stacker.logger import setup_logging;"
            "from stacker.commands import Stacker;"
            "stacker = Stacker(setup_logging=setup_logging);"
            "args = stacker.parse_args({args});"
            "stacker.configure(args);args.run(args)".format(args=str(args),
                                                            lib_path=lib_path))


class CloudFormation(RunwayModule):
    """CloudFormation (Stacker) Runway Module."""

    def run_stacker(self, command='diff'):  # pylint: disable=too-many-branches
        """Run Stacker."""
        response = {'skipped_configs': False}
        stacker_cmd = [command, "--region=%s" % self.context.env_region]

        if command == 'destroy':
            stacker_cmd.append('--force')
        elif command == 'build':
            if 'CI' in self.context.env_vars:
                stacker_cmd.append('--recreate-failed')
            else:
                stacker_cmd.append('--interactive')

        if 'DEBUG' in self.context.env_vars:
            stacker_cmd.append('--verbose')  # Increase logging if requested

        stacker_env_file = get_stacker_env_file(self.path,
                                                self.context.env_name,
                                                self.context.env_region)
        stacker_env_file_present = os.path.isfile(
            os.path.join(self.path, stacker_env_file)
        )
        if isinstance(self.options.get('environments',
                                       {}).get(self.context.env_name),
                      dict):
            for (key, val) in self.options['environments'][self.context.env_name].items():  # noqa
                stacker_cmd.extend(['-e', "%s=%s" % (key, val)])
        if stacker_env_file_present:
            stacker_cmd.append(stacker_env_file)

        with change_dir(self.path):
            # Iterate through any stacker yaml configs to deploy them in order
            # or destroy them in reverse order
            for _root, _dirs, files in os.walk(self.path):
                for name in (
                        reversed(sorted(files))
                        if command == 'destroy'
                        else sorted(files)):
                    if re.match(r"runway(\..*)?\.yml", name) or (
                            name.startswith('.')):
                        # Hidden files (e.g. .gitlab-ci.yml) or runway configs
                        # definitely aren't stacker config files
                        continue
                    if os.path.splitext(name)[1] in ['.yaml', '.yml']:
                        if not (stacker_env_file_present or self.options.get(
                                'environments',
                                {}).get(self.context.env_name)):
                            response['skipped_configs'] = True
                            LOGGER.info(
                                "Skipping stacker %s of %s; no environment "
                                "file found for this environment/region "
                                "(looking for one of \"%s\")",
                                command,
                                name,
                                ', '.join(
                                    gen_stacker_env_files(self.context.env_name,  # noqa
                                                          self.context.env_region))  # noqa
                            )
                            continue
                        ensure_stacker_compat_config(
                            os.path.join(self.path, name)
                        )
                        LOGGER.info("Running stacker %s on %s in region %s",
                                    command,
                                    name,
                                    self.context.env_region)
                        stacker_cmd_str = make_stacker_cmd_string(
                            stacker_cmd + [name],
                            get_embedded_lib_path()
                        )
                        stacker_cmd_list = [sys.executable, '-c']
                        LOGGER.debug(
                            "Stacker command being executed: %s \"%s\"",
                            ' '.join(stacker_cmd_list),
                            stacker_cmd_str
                        )
                        run_module_command(
                            cmd_list=stacker_cmd_list + [stacker_cmd_str],
                            env_vars=self.context.env_vars
                        )
                break  # only need top level files
        return response

    def plan(self):
        """Run stacker diff."""
        self.run_stacker(command='diff')

    def deploy(self):
        """Run stacker build."""
        self.run_stacker(command='build')

    def destroy(self):
        """Run stacker destroy."""
        self.run_stacker(command='destroy')



"""The gen-sample command."""
import logging
import os
import shutil
from subprocess import check_output
import sys

import cfn_flip

from .base import Base

LOGGER = logging.getLogger('runway')


def generate_sample_module(module_dir):
    """Generate skeleton sample module."""
    if os.path.isdir(module_dir):
        LOGGER.error("Error generating sample module -- directory %s "
                     "already exists!",
                     module_dir)
        sys.exit(1)
    os.mkdir(module_dir)


def generate_sample_sls_module(env_root, module_dir=None):
    """Generate skeleton Serverless sample module."""
    if module_dir is None:
        module_dir = os.path.join(env_root, 'sampleapp.sls')
    generate_sample_module(module_dir)
    for i in ['config-dev-us-east-1.json', 'handler.py', 'package.json',
              'serverless.yml']:
        shutil.copyfile(
            os.path.join(os.path.dirname(os.path.dirname(__file__)),
                         'templates',
                         'serverless',
                         i),
            os.path.join(module_dir, i),
        )
    LOGGER.info("Sample Serverless module created at %s",
                module_dir)


def generate_sample_sls_tsc_module(env_root, module_dir=None):
    """Generate skeleton Serverless TypeScript sample module."""
    if module_dir is None:
        module_dir = os.path.join(env_root, 'sampleapp.sls')
    generate_sample_module(module_dir)
    for i in ['package.json', 'serverless.yml', 'tsconfig.json',
              'webpack.config.js']:
        shutil.copyfile(
            os.path.join(os.path.dirname(os.path.dirname(__file__)),
                         'templates',
                         'sls-tsc',
                         i),
            os.path.join(module_dir, i),
        )
    os.mkdir(os.path.join(module_dir, 'src'))
    for i in ['handler.spec.ts', 'handler.ts']:
        shutil.copyfile(
            os.path.join(os.path.dirname(os.path.dirname(__file__)),
                         'templates',
                         'sls-tsc',
                         'src',
                         i),
            os.path.join(module_dir, 'src', i),
        )
    LOGGER.info("Sample Serverless TypeScript module created at %s",
                module_dir)


def generate_sample_cdk_module(env_root, module_dir=None):
    """Generate skeleton CDK sample module."""
    if module_dir is None:
        module_dir = os.path.join(env_root, 'sampleapp.cdk')
    generate_sample_module(module_dir)
    for i in ['cdk.json', 'index.ts', 'package.json', 'tsconfig.json']:
        shutil.copyfile(
            os.path.join(os.path.dirname(os.path.dirname(__file__)),
                         'templates',
                         'cdk',
                         i),
            os.path.join(module_dir, i),
        )
    LOGGER.info("Sample CDK module created at %s", module_dir)


def generate_sample_cfn_module(env_root, module_dir=None):
    """Generate skeleton CloudFormation sample module."""
    if module_dir is None:
        module_dir = os.path.join(env_root, 'sampleapp.cfn')
    generate_sample_module(module_dir)
    for i in ['stacks.yaml', 'dev-us-east-1.env']:
        shutil.copyfile(
            os.path.join(os.path.dirname(os.path.dirname(__file__)),
                         'templates',
                         'cfn',
                         i),
            os.path.join(module_dir, i)
        )
    os.mkdir(os.path.join(module_dir, 'templates'))
    with open(os.path.join(module_dir,
                           'templates',
                           'tf_state.yml'), 'w') as stream:
        stream.write(
            cfn_flip.flip(
                check_output(
                    [sys.executable,
                     os.path.join(os.path.dirname(os.path.dirname(__file__)),  # noqa
                                  'templates',
                                  'stacker',
                                  'tfstate_blueprints',
                                  'tf_state.py')]
                )

            )
        )
    LOGGER.info("Sample CloudFormation module created at %s",
                module_dir)


def generate_sample_stacker_module(env_root, module_dir=None):
    """Generate skeleton Stacker sample module."""
    if module_dir is None:
        module_dir = os.path.join(env_root,
                                  'runway-sample-tfstate.cfn')
    generate_sample_module(module_dir)
    for i in ['stacks.yaml', 'dev-us-east-1.env']:
        shutil.copyfile(
            os.path.join(os.path.dirname(os.path.dirname(__file__)),
                         'templates',
                         'stacker',
                         i),
            os.path.join(module_dir, i)
        )
    os.mkdir(os.path.join(module_dir, 'tfstate_blueprints'))
    for i in ['__init__.py', 'tf_state.py']:
        shutil.copyfile(
            os.path.join(os.path.dirname(os.path.dirname(__file__)),
                         'templates',
                         'stacker',
                         'tfstate_blueprints',
                         i),
            os.path.join(module_dir, 'tfstate_blueprints', i)
        )
    os.chmod(  # make blueprint executable
        os.path.join(module_dir, 'tfstate_blueprints', 'tf_state.py'),
        os.stat(os.path.join(module_dir,
                             'tfstate_blueprints',
                             'tf_state.py')).st_mode | 0o0111
    )
    LOGGER.info("Sample Stacker module created at %s",
                module_dir)


def generate_sample_tf_module(env_root, module_dir=None):
    """Generate skeleton Terraform sample module."""
    if module_dir is None:
        module_dir = os.path.join(env_root, 'sampleapp.tf')
    generate_sample_module(module_dir)
    for i in ['.terraform-version', 'backend-us-east-1.tfvars',
              'dev-us-east-1.tfvars', 'main.tf']:
        shutil.copyfile(
            os.path.join(os.path.dirname(os.path.dirname(__file__)),
                         'templates',
                         'terraform',
                         i),
            os.path.join(module_dir, i),
        )
    LOGGER.info("Sample Terraform app created at %s",
                module_dir)


class GenSample(Base):
    """Extend Base with execute to run the module generators."""

    def execute(self):
        """Run selected module generator."""
        if self.options['cfn']:
            generate_sample_cfn_module(self.env_root)
        elif self.options['sls']:
            generate_sample_sls_module(self.env_root)
        elif self.options['sls-tsc']:
            generate_sample_sls_tsc_module(self.env_root)
        elif self.options['stacker']:
            generate_sample_stacker_module(self.env_root)
        elif self.options['tf']:
            generate_sample_tf_module(self.env_root)
        elif self.options['cdk']:
            generate_sample_cdk_module(self.env_root)



"""runway base module."""
from __future__ import print_function

import glob
import logging
import os
from subprocess import check_call, check_output
from subprocess import CalledProcessError
import sys

import yaml

# from stacker.util import parse_cloudformation_template
# parse_cloudformation_template wraps yaml_parse; it would be better to call it
# from util but that would require sys.path shenanigans here
from ..embedded.stacker.awscli_yamlhelper import yaml_parse as parse_cloudformation_template  # noqa
from ..util import (
    change_dir, get_embedded_lib_path, ignore_exit_code_0, use_embedded_pkgs,
    which
)
from .. import __version__ as version

LOGGER = logging.getLogger('runway')


class Base(object):
    """Base class for deployer classes."""

    def __init__(self, options, env_root=None, runway_config_dir=None):
        """Initialize base class."""
        self.options = options

        if env_root is None:
            self.env_root = os.getcwd()
        else:
            self.env_root = env_root

        if runway_config_dir is None:
            self.runway_config_path = os.path.join(
                self.env_root,
                'runway.yml'
            )
        else:
            self.runway_config_path = os.path.join(
                runway_config_dir,
                'runway.yml'
            )
        self._runway_config = None

    def get_env_dirs(self):
        """Return list of directories in env_root."""
        repo_dirs = next(os.walk(self.env_root))[1]
        if '.git' in repo_dirs:
            repo_dirs.remove('.git')  # not relevant for any repo operations
        return repo_dirs

    def get_python_files_at_env_root(self):
        """Return list of python files in env_root."""
        return glob.glob(os.path.join(self.env_root, '*.py'))

    def get_yaml_files_at_env_root(self):
        """Return list of yaml files in env_root."""
        yaml_files = glob.glob(
            os.path.join(self.env_root, '*.yaml')
        )
        yml_files = glob.glob(
            os.path.join(self.env_root, '*.yml')
        )
        return yaml_files + yml_files

    def lint(self, base_dir=None, dirs_to_scan=None):
        """Call code linters."""
        from flake8.main import application as flake8_app
        from yamllint.cli import run as yamllint_run

        if base_dir is None:
            base_dir = self.env_root
        if dirs_to_scan is None:
            dirs_to_scan = self.get_env_dirs()

        if os.path.isfile(os.path.join(base_dir, '.flake8')):
            # config file in env will be picked up automatically
            flake8_config = []
        else:
            # no config file in env; use runway defaults
            flake8_config = [
                ('--append-config=' + os.path.join(
                    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),  # noqa
                    'templates',
                    '.flake8'
                ))
            ]
        if os.path.isfile(os.path.join(base_dir, '.yamllint')):
            yamllint_config = os.path.join(base_dir, '.yamllint')
        elif os.path.isfile(os.path.join(base_dir, '.yamllint.yml')):
            yamllint_config = os.path.join(base_dir, '.yamllint.yml')
        else:
            yamllint_config = os.path.join(
                os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
                'templates',
                '.yamllint.yml'
            )
        with change_dir(base_dir):
            with ignore_exit_code_0():
                LOGGER.info('Starting Flake8 linting...')
                flake8_run = flake8_app.Application()
                flake8_run.run(
                    flake8_config + dirs_to_scan +  self.get_python_files_at_env_root()  # noqa pylint: disable=line-too-long
                )
                flake8_run.exit()
            with ignore_exit_code_0():
                LOGGER.info('Flake8 linting complete.')
                LOGGER.info('Starting yamllint...')
                yamllint_run(
                    ["--config-file=%s" % yamllint_config] + dirs_to_scan + self.get_yaml_files_at_env_root()  # noqa pylint: disable=line-too-long
                )
            LOGGER.info('yamllint complete.')

    def get_cookbook_dirs(self, base_dir=None):
        """Find cookbook directories."""
        if base_dir is None:
            base_dir = self.env_root

        cookbook_dirs = []
        dirs_to_skip = set(['.git'])
        for root, dirs, files in os.walk(base_dir):  # pylint: disable=W0612
            dirs[:] = [d for d in dirs if d not in dirs_to_skip]
            for name in files:
                if name == 'metadata.rb':
                    if 'cookbook' in os.path.basename(os.path.dirname(root)):
                        cookbook_dirs.append(root)
        return cookbook_dirs

    def cookbook_tests(self, base_dir=None):
        """Run cookbook tests."""
        if base_dir is None:
            base_dir = self.env_root

        cookbook_dirs = self.get_cookbook_dirs(base_dir)

        if cookbook_dirs:
            if which('foodcritic') is None or which('cookstyle') is None:
                LOGGER.error('"foodcritic" and/or "cookstyle" not found -- '
                             'please ensure ChefDK is installed.')
                sys.exit(1)

            for path in cookbook_dirs:
                check_call(['foodcritic', '-f any', path])
                check_call(['cookstyle', '-P', path])

    def python_tests(self, base_dir=None, pylint_rc_file=None):  # noqa pylint: disable=too-many-branches,too-many-locals
        """Run python tests."""
        from pylint.lint import Run as PylintRun

        if base_dir is None:
            base_dir = self.env_root
        if pylint_rc_file is None:
            if os.path.isfile(os.path.join(base_dir, '.pylintrc')):
                pylint_config = [
                    "--rcfile=%s" % os.path.join(base_dir, '.pylintrc')
                ]
            else:
                # Only reporting on errors ('-E') overrides any ignored errors
                # set in .pylintrc, so it is only being used here when a
                # pylint configuration file is not being used.
                pylint_config = ['-E']

        # Check all python files in repo
        dirs_to_skip = set(['.git',
                            'node_modules',
                            '.serverless'])
        nonblueprint_files = []
        blueprint_files = []
        for root, dirs, files in os.walk(base_dir):
            dirs[:] = [d for d in dirs if d not in dirs_to_skip]
            for name in files:
                filepath = os.path.join(root, name)
                if name[-3:] == '.py' and (
                        root.endswith('blueprints') and
                        not filepath.endswith('__init__.py')):
                    blueprint_files.append(filepath)
                elif name[-3:] == '.py':
                    nonblueprint_files.append(filepath)

        if nonblueprint_files + blueprint_files:
            LOGGER.info("Checking python files with pylint (\"No config file "
                        "found...\" messages can be ignored)")
            with use_embedded_pkgs():  # for embedded stacker
                with ignore_exit_code_0():
                    LOGGER.debug("Executing pylint with the following options: \"%s\"",  # noqa
                                 ' '.join(pylint_config + nonblueprint_files + blueprint_files))  # noqa pylint: disable=line-too-long
                    PylintRun(pylint_config + nonblueprint_files + blueprint_files)  # noqa
            LOGGER.info('pylint complete.')
            for filepath in blueprint_files:
                try:
                    shell_out_env = os.environ.copy()
                    if 'PYTHONPATH' in shell_out_env:
                        shell_out_env['PYTHONPATH'] = (
                            "%s:%s" % (get_embedded_lib_path(),
                                       shell_out_env['PYTHONPATH'])
                        )
                    else:
                        shell_out_env['PYTHONPATH'] = get_embedded_lib_path()
                    cfn_template = check_output(
                        [sys.executable, filepath],
                        env=shell_out_env
                    ).decode()
                    if not cfn_template:
                        raise ValueError('Template output should not be empty!')  # noqa
                    parse_cloudformation_template(cfn_template)
                except:  # noqa - Bare except fine in this context
                    print("Error while checking %s for valid "
                          "YAML/JSON output" % filepath)
                    raise

    def cfn_lint(self):
        """Run cfn-lint on templates."""
        if os.path.isfile(os.path.join(self.env_root,
                                       '.cfnlintrc')):
            LOGGER.info('Starting cfn-lint checks...')
            try:
                check_call([sys.executable,
                            '-c',
                            "import sys;from cfnlint.__main__ import main;sys.argv = ['cfn-lint'];sys.exit(main())"])  # noqa pylint: disable=line-too-long
            except CalledProcessError:
                sys.exit(1)
            LOGGER.info('cfn-lint complete')
        else:
            LOGGER.debug('Skipping cfn-lint checks (no .cfnlintrc file '
                         'defined)')

    def test(self):
        """Execute tests."""
        self.lint()
        self.cookbook_tests()
        self.python_tests()
        self.cfn_lint()

    def path_only_contains_dirs(self, path):
        """Return boolean on whether a path only contains directories."""
        pathlistdir = os.listdir(path)
        if pathlistdir == []:
            return True
        if any(os.path.isfile(os.path.join(path, i)) for i in pathlistdir):
            return False
        return all(self.path_only_contains_dirs(os.path.join(path, i)) for i in pathlistdir)  # noqa

    def get_empty_dirs(self, path):
        """Return a list of empty directories in path."""
        empty_dirs = []
        for i in os.listdir(path):
            child_path = os.path.join(path, i)
            if i == '.git' or os.path.isfile(child_path) or os.path.islink(child_path):  # noqa
                continue
            if self.path_only_contains_dirs(child_path):
                empty_dirs.append(i)
        return empty_dirs

    def parse_runway_config(self):
        """Read and parse runway.yml."""
        if not os.path.isfile(self.runway_config_path):
            LOGGER.error("Runway config file was not found (looking for "
                         "%s)",
                         self.runway_config_path)
            sys.exit(1)
        with open(self.runway_config_path) as data_file:
            return yaml.safe_load(data_file)

    @property
    def runway_config(self):
        """Return parsed runway.yml."""
        if not self._runway_config:
            self._runway_config = self.parse_runway_config()
        return self._runway_config

    @staticmethod
    def version():
        """Show current package version."""
        print(version)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import re

from past.builtins import basestring
from builtins import object
from string import Template

from .exceptions import InvalidLookupCombination, UnresolvedVariable, \
    UnknownLookupType, FailedVariableLookup, FailedLookup, \
    UnresolvedVariableValue, InvalidLookupConcatenation
from .lookups.registry import LOOKUP_HANDLERS


class LookupTemplate(Template):

    """A custom string template we use to replace lookup values"""
    idpattern = r'[_a-z][^\$\{\}]*'


def resolve_variables(variables, context, provider):
    """Given a list of variables, resolve all of them.

    Args:
        variables (list of :class:`stacker.variables.Variable`): list of
            variables
        context (:class:`stacker.context.Context`): stacker context
        provider (:class:`stacker.provider.base.BaseProvider`): subclass of the
            base provider

    """
    for variable in variables:
        variable.resolve(context, provider)


class Variable(object):
    """Represents a variable passed to a stack.

    Args:
        name (str): Name of the variable
        value (any): Initial value of the variable from the config (str, list,
                     dict)
    """

    def __init__(self, name, value):
        self.name = name
        self._raw_value = value
        self._value = VariableValue.parse(value)

    @property
    def value(self):
        """Return the current value of the Variable.
        """
        try:
            return self._value.value()
        except UnresolvedVariableValue:
            raise UnresolvedVariable("<unknown>", self)
        except InvalidLookupConcatenation as e:
            raise InvalidLookupCombination(e.lookup, e.lookups, self)

    @property
    def resolved(self):
        """Boolean for whether the Variable has been resolved.

        Variables only need to be resolved if they contain lookups.
        """
        return self._value.resolved()

    def resolve(self, context, provider):
        """Recursively resolve any lookups with the Variable.

        Args:
            context (:class:`stacker.context.Context`): Current context for
                building the stack
            provider (:class:`stacker.provider.base.BaseProvider`): subclass of
                the base provider

        """
        try:
            self._value.resolve(context, provider)
        except FailedLookup as e:
            raise FailedVariableLookup(self.name, e.lookup, e.error)

    def dependencies(self):
        """
        Returns:
            Set[str]: Stack names that this variable depends on
        """
        return self._value.dependencies()


class VariableValue(object):
    """
    Abstract Syntax Tree base object to parse the value for a variable
    """
    def value(self):
        return NotImplementedError()

    def __iter__(self):
        return NotImplementedError()

    def resolved(self):
        """
        Returns:
            bool: Whether value() will not raise an error
        """
        return NotImplementedError()

    def resolve(self, context, provider):
        pass

    def dependencies(self):
        return set()

    def simplified(self):
        """
        Return a simplified version of the Value.
        This can be used to e.g. concatenate two literals in to one literal, or
        to flatten nested Concatenations

        Returns:
            VariableValue
        """
        return self

    @classmethod
    def parse(cls, input_object):
        if isinstance(input_object, list):
            return VariableValueList.parse(input_object)
        elif isinstance(input_object, dict):
            return VariableValueDict.parse(input_object)
        elif not isinstance(input_object, basestring):
            return VariableValueLiteral(input_object)
        # else:  # str

        tokens = VariableValueConcatenation([
            VariableValueLiteral(t)
            for t in re.split(r'(\$\{|\}|\s+)', input_object)
        ])

        opener = '${'
        closer = '}'

        while True:
            last_open = None
            next_close = None
            for i, t in enumerate(tokens):
                if not isinstance(t, VariableValueLiteral):
                    continue

                if t.value() == opener:
                    last_open = i
                    next_close = None
                if last_open is not None and \
                        t.value() == closer and \
                        next_close is None:
                    next_close = i

            if next_close is not None:
                lookup_data = VariableValueConcatenation(
                    tokens[(last_open + len(opener) + 1):next_close]
                )
                lookup = VariableValueLookup(
                    lookup_name=tokens[last_open + 1],
                    lookup_data=lookup_data,
                )
                tokens[last_open:(next_close + 1)] = [lookup]
            else:
                break

        tokens = tokens.simplified()

        return tokens


class VariableValueLiteral(VariableValue):
    def __init__(self, value):
        self._value = value

    def value(self):
        return self._value

    def __iter__(self):
        yield self

    def resolved(self):
        return True

    def __repr__(self):
        return "Literal<{}>".format(repr(self._value))


class VariableValueList(VariableValue, list):
    @classmethod
    def parse(cls, input_object):
        acc = [
            VariableValue.parse(obj)
            for obj in input_object
        ]
        return cls(acc)

    def value(self):
        return [
            item.value()
            for item in self
        ]

    def resolved(self):
        accumulator = True
        for item in self:
            accumulator = accumulator and item.resolved()
        return accumulator

    def __repr__(self):
        return "List[{}]".format(', '.join([repr(value) for value in self]))

    def __iter__(self):
        return list.__iter__(self)

    def resolve(self, context, provider):
        for item in self:
            item.resolve(context, provider)

    def dependencies(self):
        deps = set()
        for item in self:
            deps.update(item.dependencies())
        return deps

    def simplified(self):
        return [
            item.simplified()
            for item in self
        ]


class VariableValueDict(VariableValue, dict):
    @classmethod
    def parse(cls, input_object):
        acc = {
            k: VariableValue.parse(v)
            for k, v in input_object.items()
        }
        return cls(acc)

    def value(self):
        return {
            k: v.value()
            for k, v in self.items()
        }

    def resolved(self):
        accumulator = True
        for item in self.values():
            accumulator = accumulator and item.resolved()
        return accumulator

    def __repr__(self):
        return "Dict[{}]".format(', '.join([
            "{}={}".format(k, repr(v)) for k, v in self.items()
        ]))

    def __iter__(self):
        return dict.__iter__(self)

    def resolve(self, context, provider):
        for item in self.values():
            item.resolve(context, provider)

    def dependencies(self):
        deps = set()
        for item in self.values():
            deps.update(item.dependencies())
        return deps

    def simplified(self):
        return {
            k: v.simplified()
            for k, v in self.items()
        }


class VariableValueConcatenation(VariableValue, list):
    def value(self):
        if len(self) == 1:
            return self[0].value()

        values = []
        for value in self:
            resolved_value = value.value()
            if not isinstance(resolved_value, basestring):
                raise InvalidLookupConcatenation(value, self)
            values.append(resolved_value)
        return ''.join(values)

    def __iter__(self):
        return list.__iter__(self)

    def resolved(self):
        accumulator = True
        for item in self:
            accumulator = accumulator and item.resolved()
        return accumulator

    def __repr__(self):
        return "Concat[{}]".format(', '.join([repr(value) for value in self]))

    def resolve(self, context, provider):
        for value in self:
            value.resolve(context, provider)

    def dependencies(self):
        deps = set()
        for item in self:
            deps.update(item.dependencies())
        return deps

    def simplified(self):
        concat = []
        for item in self:
            if isinstance(item, VariableValueLiteral) and \
                    item.value() == '':
                pass

            elif isinstance(item, VariableValueLiteral) and \
                    len(concat) > 0 and \
                    isinstance(concat[-1], VariableValueLiteral):
                # Join the literals together
                concat[-1] = VariableValueLiteral(
                    concat[-1].value() + item.value()
                )

            elif isinstance(item, VariableValueConcatenation):
                # Flatten concatenations
                concat.extend(item.simplified())

            else:
                concat.append(item.simplified())

        if len(concat) == 0:
            return VariableValueLiteral('')
        elif len(concat) == 1:
            return concat[0]
        else:
            return VariableValueConcatenation(concat)


class VariableValueLookup(VariableValue):
    def __init__(self, lookup_name, lookup_data, handler=None):
        """
        Args:
            lookup_name (basestring): Name of the invoked lookup
            lookup_data (VariableValue): Data portion of the lookup
        """
        self._resolved = False
        self._value = None

        self.lookup_name = lookup_name

        if isinstance(lookup_data, basestring):
            lookup_data = VariableValueLiteral(lookup_data)
        self.lookup_data = lookup_data

        if handler is None:
            lookup_name_resolved = lookup_name.value()
            try:
                handler = LOOKUP_HANDLERS[lookup_name_resolved]
            except KeyError:
                raise UnknownLookupType(lookup_name_resolved)
        self.handler = handler

    def resolve(self, context, provider):
        self.lookup_data.resolve(context, provider)
        try:
            if type(self.handler) == type:
                # Hander is a new-style handler
                result = self.handler.handle(
                    value=self.lookup_data.value(),
                    context=context,
                    provider=provider
                )
            else:
                result = self.handler(
                    value=self.lookup_data.value(),
                    context=context,
                    provider=provider
                )
            self._resolve(result)
        except Exception as e:
            raise FailedLookup(self, e)

    def _resolve(self, value):
        self._value = value
        self._resolved = True

    def dependencies(self):
        if type(self.handler) == type:
            return self.handler.dependencies(self.lookup_data)
        else:
            return set()

    def value(self):
        if self._resolved:
            return self._value
        else:
            raise UnresolvedVariableValue(self)

    def __iter__(self):
        yield self

    def resolved(self):
        return self._resolved

    def __repr__(self):
        if self._resolved:
            return "Lookup<{r} ({t} {d})>".format(
                r=self._value,
                t=self.lookup_name,
                d=repr(self.lookup_data),
            )
        else:
            return "Lookup<{t} {d}>".format(
                t=self.lookup_name,
                d=repr(self.lookup_data),
            )

    def __str__(self):
        return "${{{type} {data}}}".format(
            type=self.lookup_name.value(),
            data=self.lookup_data.value(),
        )

    def simplified(self):
        return VariableValueLookup(
            lookup_name=self.lookup_name,
            lookup_data=self.lookup_data.simplified(),
        )



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import str
from botocore.exceptions import ClientError
import re
from stacker.session_cache import get_session

from . import LookupHandler
from ...util import read_value_from_path

TYPE_NAME = 'dynamodb'


class DynamodbLookup(LookupHandler):
    @classmethod
    def handle(cls, value, **kwargs):
        """Get a value from a dynamodb table

        dynamodb field types should be in the following format:

            [<region>:]<tablename>@<primarypartionkey>:<keyvalue>.<keyvalue>...

        Note: The region is optional, and defaults to the environment's
        `AWS_DEFAULT_REGION` if not specified.
        """
        value = read_value_from_path(value)
        table_info = None
        table_keys = None
        region = None
        table_name = None
        if '@' in value:
            table_info, table_keys = value.split('@', 1)
            if ':' in table_info:
                region, table_name = table_info.split(':', 1)
            else:
                table_name = table_info
        else:
            raise ValueError('Please make sure to include a tablename')

        if not table_name:
            raise ValueError('Please make sure to include a dynamodb table '
                             'name')

        table_lookup, table_keys = table_keys.split(':', 1)

        table_keys = table_keys.split('.')

        key_dict = _lookup_key_parse(table_keys)
        new_keys = key_dict['new_keys']
        clean_table_keys = key_dict['clean_table_keys']

        projection_expression = _build_projection_expression(clean_table_keys)

        # lookup the data from dynamodb
        dynamodb = get_session(region).client('dynamodb')
        try:
            response = dynamodb.get_item(
                TableName=table_name,
                Key={
                    table_lookup: new_keys[0]
                },
                ProjectionExpression=projection_expression
            )
        except ClientError as e:
            if e.response['Error']['Code'] == 'ResourceNotFoundException':
                raise ValueError(
                    'Cannot find the dynamodb table: {}'.format(table_name))
            elif e.response['Error']['Code'] == 'ValidationException':
                raise ValueError(
                    'No dynamodb record matched the partition key: '
                    '{}'.format(table_lookup))
            else:
                raise ValueError('The dynamodb lookup {} had an error: '
                                 '{}'.format(value, e))
        # find and return the key from the dynamo data returned
        if 'Item' in response:
            return (_get_val_from_ddb_data(response['Item'], new_keys[1:]))
        else:
            raise ValueError(
                'The dynamodb record could not be found using the following '
                'key: {}'.format(new_keys[0]))


def _lookup_key_parse(table_keys):
    """Return the order in which the stacks should be executed.

    Args:
        dependencies (dict): a dictionary where each key should be the
            fully qualified name of a stack whose value is an array of
            fully qualified stack names that the stack depends on. This is
            used to generate the order in which the stacks should be
            executed.

    Returns:
        dict: includes a dict of lookup types with data types ('new_keys')
              and a list of the lookups with without ('clean_table_keys')

    """
    # we need to parse the key lookup passed in
    regex_matcher = '\[([^\]]+)]'
    valid_dynamodb_datatypes = ['M', 'S', 'N', 'L']
    clean_table_keys = []
    new_keys = []

    for key in table_keys:
        match = re.search(regex_matcher, key)
        if match:
            # the datatypes are pulled from the dynamodb docs
            if match.group(1) in valid_dynamodb_datatypes:
                match_val = str(match.group(1))
                key = key.replace(match.group(0), '')
                new_keys.append({match_val: key})
                clean_table_keys.append(key)
            else:
                raise ValueError(
                    ('Stacker does not support looking up the datatype: {}')
                    .format(str(match.group(1))))
        else:
            new_keys.append({'S': key})
            clean_table_keys.append(key)
    key_dict = {}
    key_dict['new_keys'] = new_keys
    key_dict['clean_table_keys'] = clean_table_keys

    return key_dict


def _build_projection_expression(clean_table_keys):
    """Given cleaned up keys, this will return a projection expression for
    the dynamodb lookup.

    Args:
        clean_table_keys (dict): keys without the data types attached

    Returns:
        str: A projection expression for the dynamodb lookup.
    """
    projection_expression = ''
    for key in clean_table_keys[:-1]:
        projection_expression += ('{},').format(key)
    projection_expression += clean_table_keys[-1]
    return projection_expression


def _get_val_from_ddb_data(data, keylist):
    """Given a dictionary of dynamodb data (including the datatypes) and a
    properly structured keylist, it will return the value of the lookup

    Args:
        data (dict): the raw dynamodb data
            keylist(list): a list of keys to lookup. This must include the
                datatype

    Returns:
        various: It returns the value from the dynamodb record, and casts it
            to a matching python datatype
    """
    next_type = None
    # iterate through the keylist to find the matching key/datatype
    for k in keylist:
        for k1 in k:
            if next_type is None:
                data = data[k[k1]]
            else:
                temp_dict = data[next_type]
                data = temp_dict[k[k1]]
            next_type = k1
    if next_type == 'L':
        # if type is list, convert it to a list and return
        return _convert_ddb_list_to_list(data[next_type])
    if next_type == 'N':
        # TODO: handle various types of 'number' datatypes, (e.g. int, double)
        # if a number, convert to an int and return
        return int(data[next_type])
    # else, just assume its a string and return
    return str(data[next_type])


def _convert_ddb_list_to_list(conversion_list):
    """Given a dynamodb list, it will return a python list without the dynamodb
        datatypes

    Args:
        conversion_list (dict): a dynamodb list which includes the
            datatypes

    Returns:
        list: Returns a sanitized list without the dynamodb datatypes
    """
    ret_list = []
    for v in conversion_list:
        for v1 in v:
            ret_list.append(v[v1])
    return ret_list
